% Relaxation Voltage Calculation for 2D Electrostatic Problems
%
% Inputs:
% BoundaryConditions - The boundary condition matrix can be any dimension NxM.  
% If a location in the matrix represents a boundary condition, simply set that
% point equal to the boundary voltage.  If voltage at a location is to be calculated,
% set that equal to the NaN value.  For this type of relaxation algorithm, the boundary
% conditions around the edge of the matrix must be specified.  Other points in the
% boundary condition matrix are optional.
%
% EquipotentialLines - An array of voltage values for equipotential lines to plot in 
% the final figure.
%
% Author:  Greg Durgin  10/29/2003 (Copyright Georgia Tech)
%

function   V = PlotVoltageXY( BoundaryConditions, EquipotentialLines, MaximumIterations, Tolerance )

FAST = true;       % CHANGE THIS TO false IF YOU WANT THE ANIMATION

%-------------------------------------------------------------------
% INITIALIZATION STEPS

if ~exist('MaximumIterations')
    MaximumIterations = 1000;           % the most number of iterations for calculation
end;

if ~exist('Tolerance')
    Tolerance = 0.001;                  % criterion for convergence
end;

Mask = isnan(BoundaryConditions);       % where are all the values for calculation?
V = zeros(size(BoundaryConditions));    % initialize voltage matrix
V(~Mask) = BoundaryConditions(~Mask);   % set voltages wherever BC
sizeV = size(V);                        % save dimensions of matrix 
X = repmat(1:sizeV(2),sizeV(1),1);      % make a matrix of row references
Y = repmat((1:sizeV(1))',1,sizeV(2));   % make a matrix of column references
i = 1;                                  % start iteration counter at 1

if ~exist('EquipotentialLines')         % make default equipotential lines if not specified
    EquipotentialLines = linspace(min(min(V(~Mask))),max(max(V(~Mask))),11);
end;    

%-------------------------------------------------------------------
% RELAXATION ALGORITHM

MaxDiff = 1e9;                          % initialize to some arbitrarily large value
fprintf('\nIteration (%i Max): ', MaximumIterations);
if ~FAST
    figure(1);
end;
while i < MaximumIterations && MaxDiff > Tolerance,    % jump out if convergence (or too many iterations)
    Vold = V;                           % save old results
    ind = sub2ind(sizeV,Y(Mask),X(Mask));
    ind_up = sub2ind(sizeV,Y(Mask)-1,X(Mask));
    ind_dn = sub2ind(sizeV,Y(Mask)+1,X(Mask));
    ind_lt = sub2ind(sizeV,Y(Mask),X(Mask)-1);
    ind_rt = sub2ind(sizeV,Y(Mask),X(Mask)+1);
    V(ind) = (Vold(ind_up)+Vold(ind_dn)+Vold(ind_lt)+Vold(ind_rt))/4;
    MaxDiff = max(max(abs(V-Vold)));    % compute maximum change
    fprintf( '%i(%1.4f)..', i, MaxDiff ); % mark new iteration number
    i = i + 1;                          % advance to next iteration
    if ~FAST
        imagesc(V); axis equal; axis xy; colorbar; title('Voltage Level'); drawnow;
    end;
end;

figure(1);
imagesc(V); axis equal; axis xy; colorbar; title('Voltage Level');


%-------------------------------------------------------------------
% CONTOUR PLOTTER
warning off MATLAB:divideByZero;
figure(2);
plot( [1 1 sizeV(2) sizeV(2) 1], [1 sizeV(1) sizeV(1) 1 1], 'k:' );
axis( [0 sizeV(2)+1 0 sizeV(1)+1] );
hold on;
for v = EquipotentialLines,
    for x = 1:(sizeV(2)-1),
        for y = 1:(sizeV(1)-1),
            Vll = V(y,x);    Vlr = V(y,x+1);
            Vul = V(y+1,x);  Vur = V(y+1,x+1);
            yl = (v-Vll)/(Vul-Vll);
            yr = (v-Vlr)/(Vur-Vlr);
            xb = (v-Vll)/(Vlr-Vll);
            xt = (v-Vul)/(Vur-Vul);
            
            if xt>=0 & xt<=1
                if yl>=0 & yl<=1
                    plot([x+xt x],[y+1 y+yl],'b-');
                elseif yr>=0 & yr<=1
                    plot([x+xt x+1],[y+1 y+yr],'b-');
                elseif xb>=0 & xb<=1
                    plot([x+xb x+xt],[y y+1],'b-');                  
                end;
            end;
            if xb>=0 & xb<=1
                if yl>=0 & yl<=1
                    plot([x+xb x],[y y+yl],'b-');
                elseif yr>=0 & yr<=1
                    plot([x+xb x+1],[y y+yr],'b-');
                end;
            elseif yl>=0 & yl<=1
                if yr>=0 & yr<=1
                    plot([x x+1],[y+yl y+yr],'b-');
                end;
            end;
        end;
    end;
end;
hold off;
title('Equipotential Lines');
axis equal;